// -*- C++ -*-

#ifndef __OPTION_H
#define __OPTION_H

#include <stdio.h>

//
// class Option
//
#define	CHANGE_RANDOM	-16001
#define CHANGE_NEXT	-16002
#define CHANGE_PREV	-16003

class Option {
    static Option * feature;		// special options
    Option * next;			// with history and automatic changing
    static int nFeatures;

protected:
    int value;
    int * oldValues;
    int history;
    int * hot;
    const char * name;

    void doSave();
    void doRestore();

public:
    Option(const char * name = "", int history = 0);

    virtual void changeTo(int) {}
    virtual void changeTo(const char * to);
    virtual void changeBy(int) {}

    void changeNext() {	  save(); changeTo(CHANGE_NEXT);   }
    void changePrev() {	  save(); changeTo(CHANGE_PREV);   }
    void changeRandom() { save(); changeTo(CHANGE_RANDOM); }

    static void changeOne();
    static void changeAll();

    virtual const char * text() const = 0;
    virtual const char * text(int) const {	// needed for OptionList
	return text();
    }

    static void save();
    static void restore();

    static void save(int to);
    static void restore(int from);

    static void getHotIni();
    static void putHotIni();

    virtual int operator()() { return 0; }	// perform some action
};

// one integer value
class OptionInt : public Option {
protected:
    int maxValue;
    int minValue;
public:
    OptionInt(const char * name, int v = CHANGE_RANDOM, 
	      int maxV = 0, int minV = 0, int hist = 0) : 
	Option(name, hist),
	maxValue(maxV), minValue(minV) {
	this->changeTo(v);
    }

    virtual void changeTo(int to);
    virtual void changeBy(int by);

    virtual const char * text() const {
	static char s[128];
	sprintf(s, "%7d", value);
	return s;
    }
    virtual operator int*() {
	return &value;
    }
    virtual operator int() const {
	return value;
    }
};

// a time 
class OptionTime : public OptionInt {
public:
    OptionTime(const char * name, int v) :
	OptionInt(name, v,0) {}
    virtual void changeTo(int to) {
	OptionInt::changeTo(to);
    }
    virtual void changeTo(const char * to);
    virtual const char* text() const {
	static char s[128];
	sprintf(s, "%5.2f sec", double(value)/100.0);
	return s;
    }
};

// something to turn on or off
class OptionOnOff : public OptionInt {
public:
    OptionOnOff(const char * name, int v) : 
	OptionInt(name, v, 2) {}
    virtual void changeTo(int to) { OptionInt::changeTo(to); }
    virtual void changeTo(const char * to);
    virtual void changeBy(int by) {
	if(by)
	    value = value ? 0 : 1;
    }
    virtual const char* text() const {
	return value ? " on" : "off";
    }
};

// something to turn on or off and to disable it
class OptionAllow : public Option {
protected:
    int (*fkt)();
public:
    enum {
	off = 0,
	on = 1,
	disabled = 2
    };

    OptionAllow(const char * name, int (*f)(), int v = CHANGE_RANDOM, int hist = 0);
    
    virtual void changeTo(int to);	// does not change "allowedness"
    virtual void changeTo(const char * to);
    virtual void changeBy(int by);	// changes btw. off (disallowed), off (allowe), on

    virtual int operator()() {
	return ( (value == 1) && (fkt != NULL) ) ? (*fkt)() : 0;
    }

    virtual const char* text() const;
    virtual operator int() const {
	return value & 1;
    }
    virtual operator int*() {
	return &value;
    }
};


// a list of possibilities, each to turn on or off
class OptionListEntry {
protected:
    char * name;				/* name (short) */
    char * desc;				/* description or long name */
public:
    OptionListEntry() : name("none"), desc("") {}
    OptionListEntry(const char * n, const char * d);
    virtual ~OptionListEntry() {
	delete name; name = NULL;
	delete desc; desc = NULL;
    }
    virtual int operator()() {
	return 0;
    }

    const char * Name() const {
	return name;
    }
    const char * Desc() const {
	return desc;
    }

    friend class OptionList;
    friend class InterfaceList;
};

class OptionList : public Option {
protected:
    OptionListEntry ** entries;
    OptionOnOff ** use;
    int nEntries;
    
    
    OptionListEntry * load_feature(const char * name, char * total_name, const char * dir,
				   OptionListEntry* (*loader)(FILE *, const char *, const char *));
    void load_dir(const char * dir, const char * extension,
		  OptionListEntry * (*loader)(FILE *, const char *, const char *) );
    
public:
    OptionList(const char * n, 
	       OptionListEntry ** e = NULL, 
	       OptionOnOff ** u = NULL, 
	       int nE = 0, 
	       int hist = 1);

    virtual void changeTo(int);
    virtual void changeTo(const char * to);
    virtual void changeBy(int);

    virtual const char * text() const { 
	return name;
    }
    virtual const char * text(int i) const {
	if( (i < 0) || (i >= nEntries))
	    return "unknown";
	return entries[i]->name;
    }
    void add(OptionListEntry *, OptionOnOff * opt);
    int defined(const char * name);

    void getIniSettings();
    void putIniSettings();

    virtual int operator()() {
	if( (value < 0) || (value >= nEntries))
	    return 0;
	return (entries[value]->operator()());
    }

    OptionListEntry * current() {
	if( (value < 0) || (value >= nEntries))
	    return NULL;
	return entries[value];
    }
    OptionListEntry * operator[](int i) {
	if( (i < 0) || (i >= nEntries)) 
	    return NULL;
	return entries[i];
    }
    int currentN() const {
	return value;
    }
    const char * currentName() const {
	if( (value < 0) || (value >= nEntries))
	    return "unknown";
	return entries[value]->name;
    }
    const char * currentDesc() const {
	if( (value < 0) || (value >= nEntries))
	    return "";
	return entries[value]->desc;
    }
    int getNEntries() const { return nEntries; }

    int load(const char * searchPath[], const char * extraPath, const char * extension,
	     OptionListEntry * (* load_function)(FILE *, const char *, const char *) );

    friend class InterfaceList;
};


extern OptionList flame;
extern OptionList flameR;
extern OptionList flameG;
extern OptionList flameB;
extern OptionList screen;
extern OptionList palette;
extern OptionList pcx;
extern OptionList translate;
extern OptionList object;
extern OptionList wave;

extern OptionInt flameGeneral;
extern OptionInt massage;
extern OptionInt table;
extern OptionInt waveScale;

extern OptionOnOff lock;
extern OptionOnOff change_little;

extern OptionAllow fft;
extern OptionAllow flashlight;
extern OptionAllow mirrorHoriz;
extern OptionAllow mirrorVert;
extern OptionOnOff tileHoriz;
extern OptionOnOff tileVert;
extern OptionAllow rgb;

extern OptionOnOff options_save;		// save options (and buffer) at end
extern OptionOnOff double_load;			// allow double loading 

extern OptionInt cthugha_verbose;		// verbose level

#endif
