#include "cthugha.h"
#include "display.h"
#include "options.h"
#include "imath.h"

#include <unistd.h>

// autoconf suggests this
#if HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) strlen((dirent)->d_name)
#else
# define dirent direct
# define NAMLEN(dirent) (dirent)->d_namlen
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# if HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif




OptionOnOff double_load("double-load", 0);	// allow double loading of features


OptionListEntry::OptionListEntry(const char * n, const char * d) {
    if(n) {
	name = new char[strlen(n)+1];
	strcpy(name, n);
    } else
	name = "none";
    if(d) {
	desc = new char[strlen(d)+1];
	strcpy(desc, d);
    } else
	desc = "";
}


OptionList::OptionList(const char * n, 
		       OptionListEntry ** e, 
		       OptionOnOff ** u, 
		       int nE , 
		       int hist ) :
    Option(n, hist), entries(NULL), use(NULL), nEntries(0) {

    // do it this way to get no problems when realocating the entries
    for(int i=0; i < nE; i++)
	add(e[i], u[i]);
}


void OptionList::changeTo(int to) {
    int i;
    
    /* check for empty set */
    if ( nEntries == 0)
	return;
	
    /* check, if there is any feature we can use */
    int nouse = 1;
    for(i=0; i < nEntries; i++)
	if ( (int)(*(use[i])) ) {
	    i = nEntries;
	    nouse = 0;
	}

    if ( nouse) {		// if none in use, use the first
	value = 0;
    } else {
	switch(to) {
	case CHANGE_RANDOM:
	    do {
		to = rand() % nEntries;
	    } while( ! (int)(*(use[to])) );
	    break;
	case CHANGE_NEXT:
	    to = value;
	    do {
		to = (to + 1) % nEntries;
	    } while( ! (int)(*(use[to])) );
	    break;
	case CHANGE_PREV:
	    to = value;
	    do {
		to = (to > 0) ? to -1 : nEntries - 1;
	    } while( ! (int)(*(use[to])) );
	    break;
	default:
	    if ( to < 0)
		to = 0;
	    if ( to >= nEntries) 
		to = nEntries -1;
	    while( ! (int)(*(use[to])) )
		to = (to+1) % nEntries;
	}
	value = to;
    }

    display_start = gettime();	/* reset counter to speed test */
    display_frames = 0;

    printfv(10, "changed %s to %s\n", name, entries[value]->name);
}

void OptionList::changeTo(const char * to) {
    int nr;
    char * pos;

    /* do nothing if to is empty */
    if ( *to == '\0' ) 
	return;

    /* first try to read as number */
    nr = strtol(to, &pos, 0);
    if( pos == to ) {
	/* not a number, so try as name */
	for(nr = 0; nr < nEntries; nr ++)
	    if(strcasecmp(to, entries[nr]->name) == 0) {	// found tha name
		changeTo(nr);
		return;
	    }
	/* found no entry, use a random value */
	printfv(2, "Unknown %s entry: %s\n", name, to);
	nr = CHANGE_RANDOM;
    }

    changeTo(nr);
}

void OptionList::changeBy(int by) {

    int to = mod(value + by, nEntries);

    changeTo(to);
}

//
// add a new entry to the list
//
void OptionList::add(OptionListEntry * entry, OptionOnOff * opt) {
    
    OptionListEntry ** new_entries = new OptionListEntry*[nEntries+1];
    OptionOnOff ** new_use = new OptionOnOff*[nEntries+1];

    for(int i=0; i < nEntries; i++) {
	new_entries[i] = entries[i];
	new_use[i] = use[i];
    }
    delete [] entries;
    delete [] use;

    entries = new_entries;
    use = new_use;
    
    entries[nEntries] = entry;
    use[nEntries] = opt;
    
    nEntries ++;

    return;
}


// 
// check if an entry is already defines
//
int OptionList::defined(const char * name) {

    for(int i=0; i < nEntries; i++)
	if ( strcmp(entries[i]->name, name) == 0)
	    return 1;
    return 0;
}

//
// get all the usages from the current ini file
//
void OptionList::getIniSettings() {
    char str[512];

    for(int i=0; i < nEntries; i++) {
	strncpy(str, name, 512);
	strncat(str, ".", 512);
	strncat(str, entries[i]->name, 512);
	getini_yesno(str, (int*)(*(use[i])) );
    }
}

//
// write all the usages to the auto-ini files
//
void OptionList::putIniSettings() {
    char str[512];
    
    for(int i=0; i < nEntries; i++) {
	if ( use[i] ) {
	    strncpy(str, name, 512);
	    strncat(str, ".", 512);
	    strncat(str, entries[i]->name, 512);
	    putini(str, *(use[i]) );
	}
    }
}

/*
 * general load-function
 */

static int is_compressed(const char * name) {
    int l = strlen(name);
    if ( l < 3)
	return 0;
    if ((name[l-3] == '.') && 
	(name[l-2] == 'g') && 
	(name[l-1] == 'z') )
	return 1;
    return 0;
}

static int has_extension(const char * name, const char * extension) {
    char * pos = strstr(name, extension);
    if ( pos == NULL)
	return 0;

    /* check if extension is at end of name, or is followed by a . */
    if ( (pos[strlen(extension)]=='\0') || (pos[strlen(extension)]=='.') ) 
	return 1;


    return 0;
}

OptionListEntry* OptionList::load_feature(const char * name, char * total_name, const char * dir,
					  OptionListEntry* (* load_function)(FILE *, const char *, const char *)) {

    int compressed = is_compressed(total_name);
    FILE * file;
	    
    if ( compressed ) {	  
	/* open with 'gzip' - through pipe */
	char cmd[PATH_MAX];
	
	printfv(3,"uncompressing and ");
	
	sprintf(cmd, "gzip -cd %s", total_name);
	
	file = popen( cmd, "r");		
	
    } else {
	/* normal open - as file */
	file = fopen( total_name, "r");
    }
    
    printfv(3,"loading: %s", total_name);
    
    /* now do the loading */
    if ( file != NULL)  {
	OptionListEntry * entry;

	/* file was openen successfully  - now read it */
	if ( (entry = (*load_function)(file, name, dir)) != NULL) {
	    printfv(3," ... OK\n");
	}
	
	if ( compressed) 
	    pclose(file);	/* close pipe */
	else
	    fclose(file);	/* close file */

	return entry;

    } else {		/* file/pipe could not be opened */
	printfv(3," ... x\n");
	return NULL;
    }
}

void OptionList::load_dir(const char * dir, const char * extension, 
			  OptionListEntry* (* load_function)(FILE *, const char *, const char *)) {
    char total_name[255];
    char feat_name[255];
    DIR * directory;
    struct dirent * entry;
    OptionListEntry * fe;
    
    if ( (directory = opendir( dir )) != NULL) {	    
	while( (entry = readdir(directory))	 != NULL) {
	    if ( ! has_extension(entry->d_name, extension) ) 
		continue;
	    
	    /* create real filename */
	    strncpy( total_name, dir, 255);
	    strncat( total_name, entry->d_name, 255);
	    
	    printfv(3, "    ");

	    /* feature name only goes till first occurence of extension */
	    strncpy(feat_name, entry->d_name, 255);
	    *strstr(feat_name, extension) = '\0';
	    
	    if ( !int(double_load) && defined(feat_name)) {
		printfv(3,"already loaded: %s\n", total_name);
		return;
	    }
	
	    if( (fe=load_feature(feat_name, total_name, dir, load_function)) != NULL)
		add(fe, new OptionOnOff("", 1));

	}
	closedir(directory);
    }
}
	    

int OptionList::load(const char * searchPath[], const char * extraPath, const char * extension,
		     OptionListEntry * (* load_function)(FILE *, const char *, const char *) ) {
    int path;
    char extra_path_dir[PATH_MAX];

    /* load normal search path */
    path = 0;
    while( searchPath[path][0] != '\0') {
	load_dir(searchPath[path], extension, load_function);
	path ++;
    }
    /* load from extra path */
    if ( extra_lib_path[0] != '\0') {
	strncpy( extra_path_dir, extra_lib_path, PATH_MAX);
	strncat( extra_path_dir, extraPath, PATH_MAX);
	load_dir( extra_path_dir, extension, load_function);
    }
    
    return 0;
}
