#include "cthugha.h"
#include "sound.h"
#include "imath.h"
#include "interface.h"

#if WITH_DSP == 1

// include the right soundcard.h
#ifdef HAVE_LINUX_SOUNDCARD_H
# include <linux/soundcard.h>
#else
# include <sys/soundcard.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>

char SoundDeviceDSP::dev_dsp[PATH_MAX] = DEV_DSP;



/*
 * Initialize /dev/dsp
 */
SoundDeviceDSP::SoundDeviceDSP(int mode) : handle(0) {
    init(mode);
}

void SoundDeviceDSP::setFragment() {

    int sound_fragment = (int(sound_fragments) << 16) | int(sound_fragment_size);
    if ( ioctl(handle, SNDCTL_DSP_SETFRAGMENT, &sound_fragment) < 0 )
	printfee("ioctl: SNDCTL_DSP_SETFRAGMENT failed.");
    
    sound_fragments.changeTo(sound_fragment >> 16);
    sound_fragment_size.changeTo(sound_fragment & 0x7fff );
    
    if( (1<<sound_fragment_size)*2*int(sound_channels) < sound_bsize) {
	printfv(1, "  sound fragment size is not set big enough.\n");
    }
}

void SoundDeviceDSP::setChannels() {
    /* set stereo or mono */
    int channels = int(sound_channels) - 1;
    if ( ioctl(handle, SNDCTL_DSP_STEREO, &channels) < 0 )
	printfee("ioctl: SNDCTL_DSP_STEREO failed");
    sound_channels.changeTo(channels + 1);
}

void SoundDeviceDSP::setSampleRate() {

    /* set sample rate */
    if ( ioctl(handle, SNDCTL_DSP_SPEED, (int*)sound_sample_rate) < 0 )
	printfee("ioctl: SNDCTL_DSP_SPEED failed");
}

void SoundDeviceDSP::setFormat() {

    /* set the sound format based on the sound_sample_size */
    int sound_format;

    switch( int(sound_sample_size) ) {
    case 8:
	sound_format = AFMT_U8;
	break;
    case 16:
	sound_format = AFMT_S16_LE;
	break;
    default:
	printfe("Unsupported sound sample size %d.\n", int(sound_sample_size));
	error = 1;
	return;
    }
	
    if ( ioctl(handle, SNDCTL_DSP_SETFMT, &sound_format) < 0)
	printfee("ioctl: SNDCTL_DSP_SETFMT failed");

    switch( sound_format) {
    case AFMT_U8:
	sound_sample_size.changeTo(8);
	break;
    case AFMT_S16_LE:
	sound_sample_size.changeTo(16);
	break;
    default:
	printfe("Unknown sound format returned by SNDCTL_DSP_SETFMT %d.\n", sound_format);
	error = 1;
	return;
    }
}


void SoundDeviceDSP::init(int mode) {
    printfv(3,"  setting %s for %s...\n", dev_dsp,
		 (mode == O_RDONLY) ? "reading" :
		 (mode == O_WRONLY) ? "writing" : "");

    this->mode = mode;

    /* Get sound-device */
    if ( (handle = open(dev_dsp, mode) ) < 0) {
 	printfee("Can't open `%s' for %s.", 
		 dev_dsp,
		 (mode == O_RDONLY) ? "reading" :
		 (mode == O_WRONLY) ? "writing" : "");
	error = 1;
	return;
    }

    switch(int(sound_method)) {

    case 0: {
	/* set sound fragment size and number of fragments */
	sound_fragment_size.changeTo( ilog2(sound_bsize) - 1);

	setFragment();
	setChannels();
	setSampleRate();
	setFormat();

	break;
    }
    case 1: {
	/* fragment size of 16 bytes (2^4), two fragments (2) */
	sound_fragments.changeTo(2);
	sound_fragment_size.changeTo(4);
	
	setFragment();
	setChannels();
	setSampleRate();
	setFormat();
	
	break;
    }
    case 2: {
	// from version 0.3

	/* create a DMA-Buffer as small as possible.
	   1. create the buffer for lowerest possible sample rate (4000) and
	   mono, with maximul SUBDIVE (4). (The buffersize is calculated
	   from the sample rate and number of channels and is not changed
	   after it is set for the first time by a call to GETBLKSIZE)
	   On my machine I get a BLKSIZE of 1024 bytes
	   2. set the sample rate and nr. of channels as specified 
	*/
	int sound_div = 4;			/* blocks as small as possible */
	if ( ioctl(handle, SNDCTL_DSP_SUBDIVIDE, &sound_div) < 0 )
	    printfee("ioctl: SNDCTL_DSP_SUBDIVIDE failed.");
	
	int dummy = 0;
	if ( ioctl(handle, SNDCTL_DSP_STEREO, &dummy) < 0 ) 
	    printfee("ioctl: SNDCTL_DSP_STEREO failed.");
	
	dummy = 0;
	if ( ioctl(handle, SNDCTL_DSP_SPEED, &dummy) <0 )
	    printfee("ioctl: SNDCTL_DSP_SPEED failed.");

	int sound_blk_size;
	if ( ioctl(handle, SNDCTL_DSP_GETBLKSIZE, &sound_blk_size) < 0 )
	    printfee("ioctl: SNDCTL_DSP_GETBLKSIZE failed.");
	
	/* Set the specified values for sample rate and nr. of channels */
	int stereo = int(sound_channels) - 1;
	if ( ioctl(handle, SNDCTL_DSP_STEREO, &stereo) < 0 )
	    printfee("ioctl: SNDCTL_DSP_STEREO failed");
	sound_channels.changeTo(stereo + 1);
	
	if ( ioctl(handle, SNDCTL_DSP_SPEED, (int*)sound_sample_rate) <0 )
	    printfee("ioctl: SNDCTL_DSP_SPEED");
	
	if ( ioctl(handle, SNDCTL_DSP_SETFMT, (int*)sound_sample_size) < 0 )
	    printfee("ioctl: SNDCTL_DSP_SETFMT");
	
	if ( ioctl(handle, SNDCTL_DSP_GETBLKSIZE, &sound_blk_size) < 0 )
	    printfee("ioctl: SNDCTL_DSP_GETBLKSIZE");
	break;
    }
    default:
	printfe("Unknown sound method %d.\n", int(sound_method));
	error = 1;
	return;
    }

    printfv(4,"    sound method       : %d\n", int(sound_method) );
    printfv(4,"    sound fragments    : %d\n", int(sound_fragments) );
    printfv(4,"    sound fragment size: %d\n", int(sound_fragment_size) );
    printfv(4,"    number of channels : %d\n", int(sound_channels) );
    printfv(4,"    sample rate        : %d\n", int(sound_sample_rate) );
    printfv(4,"    sample size        : %d\n", int(sound_sample_size) );
}

void SoundDeviceDSP::update() {
    close(handle);
    init(mode);
}


/*
 * Get sound from sound-card
 *
 * Jan Kujawa <kujawa@kallisti.montana.com> 
 *   helped a lot to do this better.
 */
void SoundDeviceDSPIn::read() {

    switch(int(sound_method)) {
    case 0: {
	audio_buf_info bi;
	int nr_read = int(sound_channels) * sound_bsize / (1<<sound_fragment_size);
	unsigned char sound_buffer[(1<<sound_fragment_size) * 2 * int(sound_channels)];
	
	/* get number of available fragments */
	if( ioctl(handle, SNDCTL_DSP_GETISPACE, &bi) < 0) {
	    printfee("ioctl: SNDCTL_DSP_GETISPACE failed.");
	}
	
	/* read any extra sound data */
	for(int i=0; i < bi.fragments-nr_read; i++ )
	    ::read(handle, sound_buffer, (1<<sound_fragment_size));
	
	/* read the sound data, that will be display */
	if( ::read(handle, sound_buffer, (1<<sound_fragment_size) * nr_read) < 0) {
	    printfee("reading sound failed."); 
	}
	
	sound_convert(sound_buffer, int(sound_channels), int(sound_sample_size));
	break;
    }
    case 1: {
	unsigned char * sbuff;
	int nr_read;
	unsigned char sound_buffer[65536];
	
	/* Important information from Jan Kujawa <kujawa@kallisti.montana.com>
	   how to do this correctly */
	for(nr_read = 0, sbuff=sound_buffer; 
	    nr_read < sound_bsize*int(sound_channels); nr_read += 32) {
	    if( ::read(handle, sbuff, 16) < 0)
		printfee("sound_read < 0"); 
	    sbuff += 16;
	    if( ::read(handle, sbuff, 16) < 0)
		printfee("sound_read < 0");
	    sbuff += 16;
	}

	break;
    }
    case 2: {
	unsigned char sound_buffer[sound_bsize * int(sound_channels)];
	int nr_read = ::read(handle, sound_buffer, sound_bsize * int(sound_channels));

	if( nr_read < 0) {
	    printfee("get_sound: read < 0.");
	}

	sound_convert(sound_buffer, int(sound_channels), int(sound_sample_size));
	break;
    }
    
    }

    /* this should no not be necessary - it only causes problems */
    if(int(sound_sync))
	ioctl(handle, SNDCTL_DSP_RESET);
}

void SoundDeviceDSPOut::write(const unsigned char *data, int size) {
    ::write(handle, data, size);
}



SoundDeviceDSP::~SoundDeviceDSP() {
    close(handle);
}

#else


char SoundDeviceDSP::dev_dsp[PATH_MAX] = "";

SoundDeviceDSP::SoundDeviceDSP(int mode) : handle(0) {
    init(mode);
}

void SoundDeviceDSP::init(int) {
    printfe("DSP device was disabled at compile time.\n");
    error = 1;
}

void SoundDeviceDSP::setFragment() {}
void SoundDeviceDSP::setChannels() {}
void SoundDeviceDSP::setSampleRate() {}
void SoundDeviceDSP::setFormat() {}
void SoundDeviceDSP::update() {}
void SoundDeviceDSPIn::read() {}
void SoundDeviceDSPOut::write(const unsigned char *, int ) {}
SoundDeviceDSP::~SoundDeviceDSP() {}


#endif



