#include "cthugha.h"
#include "sound.h"

#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/stat.h>

#include <sys/types.h>
#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifndef WEXITSTATUS
# define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
#endif
#ifndef WIFEXITED
# define WIFEXITED(stat_val) (((stat_val) & 255) == 0)
#endif


#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif


char SoundDeviceFile::name[PATH_MAX];

OptionSound soundBuffer("sound-buffer", 1024);
OptionOnOff soundLoop("play-loop", 1);

enum sndfile_t{
    sndfile_raw,
    sndfile_wave
};
static enum sndfile_t sound_file_type = sndfile_raw;


#define RIFF		0x46464952	
#define WAVE		0x45564157
#define FMT		0x20746D66
#define DATA		0x61746164
#define PCM_CODE	1
#define WAVE_MONO	1
#define WAVE_STEREO	2

/* it's in chunks like .voc and AMIGA iff, but my source say there
   are in only in this combination, so I combined them in one header;
   it works on all WAVE-file I have
*/
typedef struct _waveheader {
  unsigned long	main_chunk;	/* 'RIFF' */
  unsigned long	length;		/* filelen */
  unsigned long	chunk_type;	/* 'WAVE' */

  unsigned long	sub_chunk;	/* 'fmt ' */
  unsigned long	sc_len;		/* length of sub_chunk, =16 */
  unsigned short format;	/* should be 1 for PCM-code */
  unsigned short modus;		/* 1 Mono, 2 Stereo */
  unsigned long	sample_fq;	/* frequence of sample */
  unsigned long	byte_p_sec;
  unsigned short byte_p_spl;	/* samplesize; 1 or 2 bytes */
  unsigned short bit_p_spl;	/* 8, 12 or 16 bit */ 

  unsigned long	data_chunk;	/* 'data' */
  unsigned long	data_length;	/* samplecount */
} WaveHeader;


SoundDeviceFile::SoundDeviceFile(int loud) : SoundDevice(), dsp(NULL), buffer_pid(-1) {

    if(openBuffer()) {
	error = 1;
	return;
    } else if(this->header()) {
	error = 1;
	return;
    } else if(loud)
	dsp = new SoundDeviceDSPOut;
}

int SoundDeviceFile::openBuffer() {
    const int chunk_size = 4096;
    const int buffer_size = int(soundBuffer) * 1024;

    if( buffer_size <= (4*chunk_size) ) {
	// it is not work having such a small buffer
	buffer_pid = -1;
	this->open();

    } else {
	printfv(3,"    starting sound buffer process\n");
	printfv(4,"    buffer size : %dK\n", buffer_size/1024);

	int pipedes[2];
	if( pipe(pipedes) == -1) {
	    printfee("Can not open pipe for sound buffer.");
	    return 1;
	}
	
	switch(buffer_pid = fork() ) {
	case -1:
	    printfee("Can not fork sound buffer process.");
	    return 1;
	case 0: {	// child
	    if(dsp)		// if this is initialized the second time the dsp device
		delete dsp;	// was already opened. we don't need it here.

	    ::close(pipedes[0]);	// close reading side of pipe
	    if( this->open() ) {	// open the file (or start program)
		::close(pipedes[1]);
		abort();
	    }
	    
	    unsigned char * buffer = new unsigned char[buffer_size+chunk_size];
	    
	    int buffer_pos = 0;		// start of data in the buffer (read position)
	    int buffer_fill = 0;		// number of elements in the buffer
	    
	    while(!feof(file) && !ferror(file) ){
		int r = 0;
		int w = 0;
		
		unsigned char * read_pos = buffer + (buffer_pos + buffer_fill) % buffer_size;
		unsigned char * write_pos = buffer + buffer_pos;
		
		if( buffer_fill < chunk_size) {			// we must read now
		    if(buffer_pos > 2*chunk_size)
			printfv(4, "sound buffer empty!\n");
		    r = fread(read_pos, 1, chunk_size, file);
		}
		
		else if( (buffer_fill+chunk_size) >= buffer_size)	// we must write now
		    w = write(pipedes[1], write_pos, chunk_size);
		
		else {
		    fd_set rfds, wfds;
		    FD_ZERO(&rfds); 
		    FD_ZERO(&wfds);
		    
		    FD_SET(fileno(file), &rfds);
		    FD_SET(pipedes[1], &wfds);
		    
		    switch( select(FD_SETSIZE, &rfds, &wfds, NULL, NULL) ) {
		    case -1:
			printfee("Error in select.");
			break;
		    case 0:
			break;
		    default:
			if( FD_ISSET(fileno(file), &rfds) ) // reading possible
			    r = fread(read_pos, 1, chunk_size, file);
			if( FD_ISSET(pipedes[1], &wfds) )	// write possible
			    w = write(pipedes[1], write_pos, chunk_size);
		    }
		}
		
		// writing was not possible
		if(w == -1) {
		    ::close(pipedes[1]);
		    this->close();
		    printfv(3, "    sound buffer process %d stopped.", getpid());
		    abort();
		}
		
		// check for read over the end of the buffer
		if( (r > 0) && (((buffer_pos + buffer_fill)%buffer_size + r) > buffer_size) ) {
		    memcpy(buffer, buffer+buffer_size, 
			   (buffer_pos + buffer_fill)%buffer_size + r - buffer_size);
		}
		
		// update buffer position and size
		buffer_pos = (buffer_pos + w) % buffer_size;
		buffer_fill = buffer_fill + r - w;
	    }
	    
	    // write the remaining data
	    if(buffer_fill > 0) {
		int n1 = (buffer_pos + buffer_fill) - buffer_size;
		if(n1 > 0) {
		    write(pipedes[1], buffer + buffer_pos, buffer_size - buffer_pos);
		    write(pipedes[1], buffer, n1);
		} else
		    write(pipedes[1], buffer + buffer_pos, buffer_fill);
	    }
	    
	    ::close(pipedes[1]);
	    this->close();
	    
	    printfv(3, "    sound buffer process %d stopped.\n", getpid());
	    abort();
	}
	default:	// parent
	    ::close(pipedes[1]);	// close writing side of pipe
	    if( (file = fdopen(pipedes[0], "r") ) == NULL) {
		printfee("Can not associate buffer pipe to file.\n");
		return 1;
	    }
	}
    }
    return 0;
}

int SoundDeviceFile::closeBuffer() {

    if(buffer_pid >= 0) {
	if(file)				// close the file, this will also cause the
	    fclose(file);			//  buffer process to stop

	if(waitpid(buffer_pid, NULL, WNOHANG) == -1) {
	    sleep(1);
	    if(waitpid(buffer_pid, NULL, WNOHANG) == -1) {
		printfv(3,"    stopping sound buffer process\n");
		
		kill(SIGTERM, buffer_pid);
		waitpid(buffer_pid, NULL, 0);
	    }
	}
    } else {
	this->close();
    }

    return 0;
}

int SoundDeviceFile::open() {
    if( (file = fopen(name, "r")) == NULL) {
	printfee("Can not open sound file `%s' for reading.", name);
	return 1;
    }
    return 0;
}

int SoundDeviceFile::header() {
    WaveHeader header;
    
    if( fread(&header, sizeof(header), 1, file) <= 0) {
	if(feof(file) )
	    printfe("Can not read header (end of file).\n");
	else
	    printfee("Can not read header.");
    }

    /* is it a .wav file */
    if( (header.main_chunk == RIFF) && 
	(header.chunk_type == WAVE) ) {
	
	sound_channels.changeTo(header.modus);
	sound_sample_size.changeTo(header.bit_p_spl);
	sound_sample_rate.changeTo(header.sample_fq);

	sound_file_type = sndfile_wave;
    } else {
	printfv(2, "  Unknown sound file format. Using raw data\n");
	sound_file_type = sndfile_raw;
    }
    return 0;
}

int SoundDeviceFile::restart() {

    closeBuffer();
    
    if(openBuffer()) {			// recreate the buffer
	return 1; 
    } else if(this->header()) {		// reread header
	return 1;
    }

    if(dsp)
	dsp->update();
    sound_communicate(0);
    return 0;
}

void SoundDeviceFile::read() {
    int size = sound_bsize * int(sound_channels) * ((int(sound_sample_size)+7)/8);
    unsigned char sound_buffer[size];
    
    /* read sound from file (through buffer) */
    size = fread(sound_buffer, 1, size, file);

    /* Play the sound */
    if(dsp)
	dsp->write(sound_buffer, size);

    if(feof(file))
	this->restart();

    sound_convert(sound_buffer, int(sound_channels), int(sound_sample_size));
}

int SoundDeviceFile::close() {
    if(file)
	if( fclose(file) ) {
	    printfee("closing %s failed.", name);
	    return 1;
	}
    return 0;
}

SoundDeviceFile::~SoundDeviceFile() {
    closeBuffer();
    if(dsp)
	delete dsp;
}






char SoundDeviceExec::fifo[PATH_MAX] = "/tmp/cthugha.com";


SoundDeviceExec::SoundDeviceExec(int loud) : childPid(-1) {
    dsp = NULL;
    buffer_pid = -1;

    if(openBuffer()) {
	error = 1;
	return;
    } else if(this->header()) {
	error = 1;
	return;
    } else if(loud)
	dsp = new SoundDeviceDSPOut;

}
int SoundDeviceExec::open() {
    
    // make a named pipe (FIFO file)
    // ignore errors, the file might already exist
    mkfifo(fifo, 0666);

    if(name[0] == '\0') {
	// now open this pipe for reading
	printfv(3, "    opening %s\n", fifo);
	if( (file = fopen(fifo, "r")) == NULL) {
	    printfee("Can not open communication pipe `%s'.\n", fifo);
	    return 1;
	}

	childPid = -1;
    } else {	
        switch( childPid = fork()) {
	case -1:
	    printfee("Can not fork for sound reading program.\n");
	    error = 1;
	    return 1;
	case 0:	
	    // child
	    // the program should open the pipe for writing
	    printfv(3, "    starting sound reader `%s'.\n", name);
	    system(name);
	    printfv(0, "    sound reader stopped.\n");
	    abort();
	    break;
	default:
	    // now open this pipe for reading
	    printfv(3, "    opening %s\n", fifo);
	    if( (file = fopen(fifo, "r")) == NULL) {
		printfee("Can not open communication pipe `%s'.\n", fifo);
		return 1;
	    }
	}
    }
    return 0;
}

int SoundDeviceExec::close() {

    if(childPid >= 0) {
	if(waitpid(childPid, NULL, WNOHANG) == -1) {
	    printfv(3,"    stopping sound reader\n");
	    
	    kill(SIGKILL, childPid);
	    waitpid(childPid, NULL, 0);
	}
    }

    return SoundDeviceFile::close();
}



int SoundDeviceExec::restart() {

    closeBuffer();

    if( !int(soundLoop)) {		// not playing again ->
	name[0] = '\0';			// set to listening mode
    }

    if(openBuffer()) {			// recreate the buffer
	return 1; 
    } else if(this->header()) {		// reread header
	return 1;
    }

    if(dsp)
	dsp->update();
    sound_communicate(0);

    return 0;
}

SoundDeviceExec::~SoundDeviceExec() {
    closeBuffer();

    if(dsp)
	delete dsp;

    unlink(fifo);
}

void SoundDeviceFile::update() {
    if(dsp) 
	dsp->update();
}
