/* 
 * Based on README.aztcd from the linux-sources (Tiny Audio CD Player)
 *       and CDPlay 0.31 from Sariel Har-Peled
 * There is not much left from the original code.
 */

#include "cthugha.h"
#include "cd_player.h"
#include "interface.h"

#if WITH_CDROM == 1

#include "keys.h"
#include "display.h"
#include "action.h"
#include "sound.h"

#include <ctype.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>

#ifdef HAVE_LINUX_CDROM_H
# include <linux/cdrom.h>
#else
# include <sys/cdrom.h>
#endif

char dev_cd[PATH_MAX] = DEV_CDROM;

OptionInt cd_first_track("cd-first-track", 0);		// start with track nr 
OptionOnOff cd_loop("cd-loop", 0);			// restart when all tracks played
OptionOnOff cd_randomplay("cd-random-play", 0);		// play random
OptionOnOff cd_eject_on_end("cd-eject-on-end", 0);	// eject CD if stoped
OptionOnOff cd_stop_on_exit("cd-stop-on-exit", 0);	// stop playing when leaving program



static struct cdrom_subchnl  subchnl; 
static int cd_playing = 0;			/* already playing ? */
static int first, last;
  
static int cd_handle = -1;				/* handle of /dev/cdrom */
static int cd_in_pause = 0;				/* now in pause-mode */

static int *random_played;
static int ejected=0;
static int no_disk=0;
static int stop_track=-1;

//
// Initialization. 
// If cd_first_track is > 0 then start playing at the specified track.
//
int init_cd() {

    cd_handle=open(dev_cd, O_RDONLY);
  
    if (cd_handle <= 0) {
	printfee("CD: already playing, no audio disk, door open");
	no_disk = 1;
	return 0;
    }
	
    cd_subchnl();			/* get information from CD */

    if ( int(cd_first_track) > 0) {
	if ( int(cd_randomplay) ) {
	    printfv(1,"  Starting random play of CD\n");
	    cd_next(0);
	} else {
	    printfv(1,"  Starting CD at track: %d.\n", int(cd_first_track) );
	    cd_track( int(cd_first_track) );
	}
    } else if(cd_playing == 0) {
	close(cd_handle);
	cd_handle = -1;
    }

    if ( int(cd_loop) )
	cd_eject_on_end.changeTo(0);

    return 0;
}

/* 
 * Close the CD-Player. 
 * Stop a possibly running CD if that is desired.
 */
int exit_cd() {
    if ( cd_handle == -1)		/* not initialized */
	return 0;

    if ( int(cd_stop_on_exit) )
	cd_stop();
		
    if (close(cd_handle)) {
	printfe("Can not close CD.\n");
	return 1;
    }
    return 0;
}

/*
 *  CD-player check for complete routine
 */
int cd_check() {

    if(!cd_playing)
	return 0;

    cd_subchnl();			/* get information from CD */
	
    /* Test wether CD-ROM has completed operation, or if the desired
       track has been reached. Not all CD-ROMs give the CDROM_AUDIO_COMPL.
       status (like mine)
    */
    if ( (subchnl.cdsc_audiostatus == CDROM_AUDIO_COMPLETED) /* 'nice' CD */
	|| (subchnl.cdsc_trk >= stop_track)	/* reached next track */
	|| ( (subchnl.cdsc_trk == 0)		/* again from start, */
	    && (stop_track > last) ) ) {	/*  while plaing last track*/
       	
	if(int(cd_randomplay))
	    cd_next(0);
	else if( int(cd_loop) )
	    cd_track(1);
	else if(int(cd_eject_on_end) && !ejected)
	    cd_eject();
	else
	    cd_stop();
    } else {
	switch( subchnl.cdsc_audiostatus ) {
	case CDROM_AUDIO_PLAY:
	    cd_playing = 1;
	    break;
	case CDROM_AUDIO_PAUSED:
	    cd_in_pause = 1;
	    break;
	default:
	    cd_playing = 0;
	    cd_in_pause = 0;
	}
    }
	    
    return 0;
}

/*****************************************************************************
 *  Primitives to access the CD:
 */

int rand_track(void) {
    int num,loop=0;
    
    if(random_played == NULL) {
	random_played = new int[last];
	for(int i=0; i < last; i++)
	    random_played[i] = 0;
    }

    num=rand()%last;
    while(random_played[num]){
	if(++num >= last){
	    num=first-1;
	    loop++;
	}
	if(loop>2)
	    break;
    }
    return loop>2?last+1:num+1;
}

#define CD_INIT    if ( cd_handle == -1) if( init_cd() ) return 1;


/* eject the CD */
int cd_eject() {
    int ret;
    CD_INIT

    if(random_played) {
	delete random_played;
	random_played = NULL;
    }
    cd_playing=0;

    if(ejected) {
	no_disk=ejected=0;
	ret = ioctl(cd_handle,CDROMSTART);	/* close the door ;) */
	if ( !ret && int(cd_first_track) > 0) {
	    if(!int(cd_randomplay))
		cd_track( int(cd_first_track) );
	    else
		cd_next(0);
	}
    }else{
	cd_stop();
	ret = ioctl(cd_handle,CDROMEJECT);
	ejected = ! ret;
    }
    return ret;
}

/* pause CD */
int cd_pause() {
	
    if(ejected || no_disk)
	return 0;

    if (!cd_playing || (cd_handle == -1))
	return 0;			/* ignore if not started playing */

    cd_subchnl();

    if (subchnl.cdsc_audiostatus == CDROM_AUDIO_PAUSED ) {
	cd_in_pause = 0;
	if (ioctl(cd_handle,CDROMRESUME)) 
	    return 1;
    } else {
       	cd_in_pause = 1;
	if (ioctl(cd_handle,CDROMPAUSE)) 
	    return 1;
    }
    return 0;
}

/* stop CD */
int cd_stop() {
    CD_INIT;

    if(ejected || no_disk)
	return 0;

    cd_playing = 0;

    ioctl(cd_handle, CDROMSTOP);

    cd_subchnl();			/* get information from CD */

    return 0;
}

int cd_readtoc() {
    struct cdrom_tochdr tocHdr;
    struct cdrom_tocentry entry;
    struct cdrom_msf      msf; 

    if(ejected || no_disk)
	return 0;

    if (ioctl(cd_handle, CDROMREADTOCHDR, &tocHdr)) {
	printfee("Can't read TOC");		// propably no disk in drive
	no_disk = 1;
	return 1;
    }
    first=tocHdr.cdth_trk0;
    last= tocHdr.cdth_trk1;
    if ((first==0)||(first>last)) {
	printfe("Illegal TOC\n");
	return 1;
    } 
    /* set end-time in msf for cd_fast() */
    entry.cdte_track = CDROM_LEADOUT;
    entry.cdte_format = CDROM_MSF;
    if (ioctl (cd_handle, CDROMREADTOCENTRY, &entry)) {
	printfee("Can't read TOC entry");
	return 1;
    }
    msf.cdmsf_min1   = entry.cdte_addr.msf.minute;
    msf.cdmsf_sec1   = entry.cdte_addr.msf.second;
    msf.cdmsf_frame1 = entry.cdte_addr.msf.frame;

    return 0;
}

/* 
 * start play starting at track 
 */
int cd_track(int track) {
    struct cdrom_tocentry entry;
    struct cdrom_msf      msf; 
    
    if(ejected || no_disk)
	return 0;

    track --;

    cd_stop();
		
    cd_readtoc();

    if (track < first)
	track=first;
    if (track > last) 
	track=last;

    entry.cdte_track = track;
    entry.cdte_format = CDROM_MSF;
    if (ioctl (cd_handle, CDROMREADTOCENTRY, &entry)) {
	printfee("Can't read TOC\n");
	return 1;
    }

    /* subtract the skip-value from the current position */
    msf.cdmsf_min0   = entry.cdte_addr.msf.minute;
    msf.cdmsf_sec0   = entry.cdte_addr.msf.second;
    msf.cdmsf_frame0 = entry.cdte_addr.msf.frame;

    /* set stop_track */
    stop_track = last + 1;
    
    /* set end-time to end of CD */
    entry.cdte_track = CDROM_LEADOUT;
    entry.cdte_format = CDROM_MSF;
    if (ioctl (cd_handle, CDROMREADTOCENTRY, &entry)) {
	printfe("Can't read TOC\n");
	return 1;
    }
    msf.cdmsf_min1   = entry.cdte_addr.msf.minute;
    msf.cdmsf_sec1   = entry.cdte_addr.msf.second;
    msf.cdmsf_frame1 = entry.cdte_addr.msf.frame;

    if( ioctl(cd_handle, CDROMSTART)) {
	printfee("Can't start CD");
	return 1;
    }
    if (ioctl(cd_handle,CDROMPLAYMSF,&msf)) {
	printfe("Can't play CD.\n");
	return 1;
    }
    cd_playing=1;

    return 0;
}

/*
 *  Read information from CD-ROM (audio-status, trk, addr)
 */
int cd_subchnl() {

    if ( cd_handle == -1)
	if(init_cd())
	    return 0;
    if ( ! cd_playing )
	cd_readtoc();

    subchnl.cdsc_format=CDROM_MSF;
    if (ioctl(cd_handle,CDROMSUBCHNL,&subchnl)) {
	subchnl.cdsc_audiostatus=CDROM_AUDIO_NO_STATUS;
	if(ejected)
	    return 0;
	else if(no_disk)
	    return 0;
	printfee("Can't SUBCHNL CD-Rom.");
	no_disk=1;
	cd_playing=0;
	return 1;
    }
    if(ejected || no_disk){
	no_disk=ejected=0;
	if( int(cd_first_track) >= 0)
	    if(!int(cd_randomplay))
		cd_track( int(cd_first_track) );
	    else
		cd_next(0);
    }

    if( (subchnl.cdsc_audiostatus != CDROM_AUDIO_PLAY) &&
       (subchnl.cdsc_audiostatus != CDROM_AUDIO_PAUSED) )
	subchnl.cdsc_trk = 0;
    else if( !cd_playing)
	cd_playing++;
	
    return 0;
}

/* go to next track */
int cd_next(int skip) {
    int track;
    struct cdrom_tocentry entry;
    struct cdrom_msf      msf; 

    CD_INIT;

    if(ejected || no_disk)
	return 0;

    if (!cd_playing++) {		/* CD not playing */
	cd_readtoc();				/* get table of contents */

	track=int(cd_randomplay) ? rand_track()	/* random track */
	    : first;				/* start at first track */

	if(int(cd_randomplay)) {
	    if ( random_played)
		free( random_played);
	    random_played=(int*)calloc(last-first+1,sizeof(int));
	}
    } else {				/* CD is already playing */
	cd_subchnl();				/* get information from CD */

	track = int(cd_randomplay) ? rand_track() 	/* next random track  */
	    : (subchnl.cdsc_trk + skip);	/* increment or drecrement */

	cd_stop();
    }
    if (int(cd_randomplay)) {
	if(track>last)
	    if( int(cd_loop)) {
		register z;
		for(z=0;z<last;z++)
		    random_played[z]=0;
		track=rand_track();				
	    }
	    else if(int(cd_eject_on_end)){
		cd_eject();
		return 0;
	    }
	    else{
		cd_stop();
		return 0;
	    }
	
	random_played[track-1]=1;
    }
    /* Check for first and last track */
    if ( track > last) 
	track=first;
    if ( track < first)
	track=last;

    entry.cdte_track = track;
    entry.cdte_format = CDROM_MSF;
    if (ioctl (cd_handle, CDROMREADTOCENTRY, &entry)) {
	printfe("Can't read TOC entry\n");
	return 1;
    }

    /* subtract the skip-value from the current position */
    msf.cdmsf_min0   = entry.cdte_addr.msf.minute;
    msf.cdmsf_sec0   = entry.cdte_addr.msf.second;
    msf.cdmsf_frame0 = entry.cdte_addr.msf.frame;

    entry.cdte_format = CDROM_MSF;
    stop_track = entry.cdte_track = 
    	( !int(cd_randomplay) || track==last ) ?CDROM_LEADOUT:track+1;
    if (ioctl (cd_handle, CDROMREADTOCENTRY, &entry)) {
	printfe("Can't read TOC entry\n");
	return 1;
    }
    msf.cdmsf_min1   = entry.cdte_addr.msf.minute;
    msf.cdmsf_sec1   = entry.cdte_addr.msf.second;
    msf.cdmsf_frame1 = entry.cdte_addr.msf.frame;

    if( ioctl(cd_handle, CDROMSTART)) {
	printfee("Can't start CD");
	return 1;
    }

    printf("Playing from %d:%d to %d:%d\n", 
	   msf.cdmsf_min0,msf.cdmsf_sec0,
	   msf.cdmsf_min1,msf.cdmsf_sec1);

    if (ioctl(cd_handle,CDROMPLAYMSF,&msf))
	printf("Drive Error\n");
    cd_playing=1;
    
    return 0;
}

int cd_fast(int skip) {
    int pos;
    struct cdrom_msf      msf; 

    CD_INIT;

    cd_subchnl();			/* get information from CD */

    if(ejected || no_disk)
	return 0;

    /* calculate the current position in frames */
    pos = subchnl.cdsc_absaddr.msf.minute * CD_SECS * CD_FRAMES +
	  subchnl.cdsc_absaddr.msf.second * CD_FRAMES +
	  subchnl.cdsc_absaddr.msf.frame;

    /* Skip by seconds */
    pos += skip * CD_FRAMES;
    if ( pos < 0)			/* play first if a begin */
	return int(cd_randomplay)?cd_next(0):cd_track(0);

    /* subtract the skip-value from the current position */
    msf.cdmsf_min0   = pos / (CD_SECS*CD_FRAMES);
    msf.cdmsf_sec0   = (pos / CD_FRAMES) % CD_SECS;
    msf.cdmsf_frame0 = pos % CD_FRAMES;

    /* we don't set end values - it is already set before, but check */
    if(((msf.cdmsf_min1 * 60 + msf.cdmsf_sec1)*CD_FRAMES+msf.cdmsf_frame1) 
       < pos ){
	printfe("Can't skip, CD untouched");
	return 0;
    }
    /* Play it */
    if( ioctl(cd_handle, CDROMSTART)) {
	printfee("Can't start CD");
	return 1;
    }
    if (ioctl(cd_handle,CDROMPLAYMSF,&msf)) { 
	printfee("Drive error or invalid address\n");
    }

    cd_subchnl();			/* get information from CD */
    return 0;
}

/*
int cd_volume(int volume) {
struct cdrom_volctrl  volctrl;
    cmd=CDROMVOLCTRL;
    printf("--Channel 0 Left  (0-255): ");
    scanf("%d",&arg1);
    printf("--Channel 1 Right (0-255): ");
    scanf("%d",&arg2);
    volctrl.channel0=arg1;
    volctrl.channel1=arg2;
    volctrl.channel2=0;
    volctrl.channel3=0;
    if (ioctl(handle,cmd,&volctrl)) 
	{ printf("Drive error or unsupported command\n");
      }
    break;  
}
*/




//
// the user interface
//
class InterfaceElementCDStatus : public InterfaceElement {
public:
    InterfaceElementCDStatus() :
	InterfaceElement("", 0,0,0) {}

    void display(int y) {
	char str[128];

	if ( subchnl.cdsc_audiostatus == CDROM_AUDIO_PLAY )
	    sprintf(str, 
		    "CD-Player Track: %d %d:%02d (%d - %d) %d:%02d",
		    subchnl.cdsc_trk,
		    subchnl.cdsc_reladdr.msf.minute,
		    subchnl.cdsc_reladdr.msf.second,
		    first, last,
		    subchnl.cdsc_absaddr.msf.minute,
		    subchnl.cdsc_absaddr.msf.second);
	else if ( subchnl.cdsc_audiostatus == CDROM_AUDIO_PAUSED )
	    sprintf(str, 
		    "CD-Player Track: %d %d:%02d (%d - %d) %d:%02d pause",
		    subchnl.cdsc_trk,
		    subchnl.cdsc_reladdr.msf.minute,
		    subchnl.cdsc_reladdr.msf.second,
		    first, last,
		    subchnl.cdsc_absaddr.msf.minute,
		    subchnl.cdsc_absaddr.msf.second);
	else
	    if(ejected)
		sprintf(str,"CD-Player: ejected");
	    else if (no_disk)
		sprintf(str,"CD-Player: no disk in drive");
	    else
		sprintf(str,"CD-Player Track: stopped");

	display_print(str, y, 'l', TEXT_COLOR_NORMAL);
    }
};


class InterfaceCD : public Interface {
    int skip;
    int fast;
public:
    InterfaceCD(InterfaceElement ** el, int nEl) :
	Interface(el, nEl) {}

    int do_key(int key) {
	switch(key) {
	case '9':    skip ++;		return 0;
	case '7':    skip --;		return 0;
	case '6':    fast += 15;	return 0;
	case '4':    fast -= 15;	return 0;
	case '8':    cd_pause();	return 0;
	case '5':    cd_stop();		return 0;
	case '2':    cd_eject();	return 0;
	default:
	    return Interface::do_key(key);
	}
    }

    void run() {
	skip = 0;
	fast = 0;

	Interface::run();

	if ( skip != 0) {
	    cd_next(skip);
	    skip = 0;
	}
	if ( fast != 0) {
	    cd_fast(fast);
	    fast = 0;
	}
    }
};


static InterfaceElement * elementsCD[] = {
    new InterfaceElementTitle("CD Player"),
    new InterfaceElementText ("----------------------------------------"),
    new InterfaceElementCDStatus(),
    new InterfaceElementText("<- prev(7)  pause(8)  -> next(9)"),
    new InterfaceElementText("<< rev(4)   stop(5)   >> fwd(6) "),
    new InterfaceElementText("           close/eject(2)       "),
    new InterfaceElementText(""),
    new InterfaceElementOption("Autostart trak: %10s", &cd_first_track),
    new InterfaceElementOption("Loop CD       : %10s", &cd_loop),
    new InterfaceElementOption("Random Play   : %10s", &cd_randomplay),
    new InterfaceElementOption("Eject on end  : %10s", &cd_eject_on_end),
    new InterfaceElementOption("Stop at exit  : %10s", &cd_stop_on_exit)
};
Interface * interfaceCD = new InterfaceCD(elementsCD, 
					   sizeof(elementsCD)/sizeof(InterfaceElement*) );


#else

//
// CDROM disabled
//

char dev_cd[PATH_MAX] = "";

int init_cd() {
    return 0;
}

int exit_cd() {
    return 0;
}

int cd_check() {
    return 0;
}


static InterfaceElement * elementsCD[] = {
    new InterfaceElementTitle("CD was disabled at compile time"),
    new InterfaceElementText("----------------------------------------")
};
Interface * interfaceCD = new Interface(elementsCD, 2);

#endif
