/*
    cthugha.c

    General functions
*/

#include "cthugha.h"
#include "information.h"			/* title, credits, ... */
#include "display.h"
#include "sound.h"
#include "action.h"				/* display_sound() */
#include "translate.h"
#include "options.h"
#include "keys.h"
#include "cd_player.h"
#include "server.h"
#include "imath.h"
#include "waves.h"
#include "Option.h"

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include <unistd.h>
#include <signal.h>
#include <errno.h>

#ifdef HAVE_VPRINTF
# include <stdarg.h>
#endif


//
// usefull to find memory errors
//
#if 0
void * operator new(size_t s) {
    void * r = malloc(s);
    printf("new %d: %x\n", s, r);
    return r;
}
void operator delete(void *d) {
    printf("delete  %x\n", d);
    if(d != NULL)
	free(d);
}
#endif



OptionInt cthugha_verbose("verbose", 1, 0, -1);		// verbosity level (no max, min=-1)


//
// print a verbose message
//
int printfv(int lvl, const char * fmt, ...) {
    if(lvl <= int(cthugha_verbose)) {
#ifdef HAVE_VPRINTF
	va_list ap;
	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);
	fflush(stdout);							
#else
	printf(fmt);
#endif
    }
    return 0;
}

//
// print an error message and the error number
//
int printfee(const char * fmt, ...) {
   printf("\n"); 
   fflush(stdout);
   
#ifdef HAVE_VPRINTF
   va_list ap;
   va_start(ap, fmt);
   vfprintf(stderr, fmt, ap);
   va_end(ap);
#else
   fprintf(stderr, fmt);
#endif

#ifdef HAVE_STRERROR
   fprintf(stderr, " (%d - %s)\n", errno, strerror(errno));
#else
   fprintf(stderr, " (%d)\n", errno);
#endif

   fflush(stderr);

   return 0;
}

//
// print an error message
//
int printfe(const char * fmt, ...) {
   printf("\n"); 
   fflush(stdout);
   
#ifdef HAVE_VPRINTF
   va_list ap;
   va_start(ap, fmt);
   vfprintf(stderr, fmt, ap);
   va_end(ap);
#else
   fprintf(stderr, fmt);
#endif

   fflush(stderr);

   return 0;
}



int cthugha_close = 0;				// closing right now
int cthugha_pause = 0;				// going to pause (^Z)

int init_startup();
int exit_startup();

int notServer = 1;

static int TRUE = 1;
int cth_init_level = -1;
struct init_exit_t {
    int * depend;
    int (*init)();
    int (*exit)();
    char * msg;
} init_exit[] = {
    {&ncurses_use,	init_ncurses,		exit_ncurses,	"ncurses"},
    {server,		init_server,		exit_server,	"sound server"},
    {&notServer,	init_flames,		NULL,		"flame functions"},
    {&TRUE,		init_display,		exit_display,	"display"},
    {&notServer,	init_translate,		NULL,		"translation tables"},
    {&notServer,	init_wave,		NULL,		"waves"},
    {&TRUE,		init_sound,		exit_sound,	"sound"},
    {&TRUE,		init_cd,		exit_cd,	"CD"},
    {&TRUE,		init_mixer,		NULL,		"mixer"},
    {&notServer,	init_sound_process,	NULL,		NULL},
    {&notServer,	read_ini_usage,		NULL,		NULL},
    {&TRUE,		init_startup,		exit_startup,	NULL}
};
int nr_init_exit = sizeof(init_exit) / sizeof(struct init_exit_t);
	

/*
 * shutdown code for cthugha
 */
void exit_cthugha() {

    for(cth_init_level --; cth_init_level >= 0; cth_init_level --) { 
	if( *(init_exit[cth_init_level].depend) ) {
	    if(init_exit[cth_init_level].exit) {
		if( init_exit[cth_init_level].msg )
		    printfv(1, "Closing %s...\n", init_exit[cth_init_level].msg);
		init_exit[cth_init_level].exit();
	    }
	}
    }
}



/*
 * get the 1/100 sec since program start
 */
int gettime() {
    struct timeval tv;
    static int starttime = 0;
    
    gettimeofday(&tv, NULL);

    if(starttime == 0) {
	starttime = tv.tv_sec;
    }
    tv.tv_sec -= starttime;

    return tv.tv_sec * 100L + tv.tv_usec / 10000L;
}


/*
 * handle ^Z and continue 
 */
void sig_tty_cont(int);
void sig_tty_stop(int) {
    printfv(0, "Stopping...\n");

    signal(SIGCONT, sig_tty_cont);		/* set, how to continue */

    cthugha_pause = 1;				/* in interface we will really stop */
}
void sig_tty_cont(int) {
    printfv(0, "Continuing...\n");

    init_graph_mode();				/* turn back on graphics */
    init_sound();				/* and sound */

    signal(SIGTSTP, sig_tty_stop);		/* set, how to stop again */

    raise(SIGCONT);				/* default action */
}
	

/*
 *
 */
int main(int argc, char * argv[]) {
    srand(time(0));			/* initialize random generator */
    atexit(exit_cthugha);		// make sure we exit correctly

    if( get_pre_params(argc, argv) )	// handle some special arguments (verbose, ...)
	return 1;

    if( cth_init(&argc, argv) )		/* special initialization */
	return 1;

    if( get_params(argc, argv) )	/* parse cmd-line and read ini-files*/
	return 1;

    title();				/* Display titlemessage */

    /* Initialization */
    init_imath();
    for(cth_init_level=0; cth_init_level < nr_init_exit; cth_init_level++) {
	if( *(init_exit[cth_init_level].depend) ) {
	    if( init_exit[cth_init_level].msg )
		printfv(1, "Initializing %s...\n", init_exit[cth_init_level].msg);
	    if( (init_exit[cth_init_level].init)() )
		return 1;
	}
    }

    signal(SIGTSTP, sig_tty_stop);	/* react to ^Z */

    cth_main();
    
    return 0;
}










