// -*-c++-*-
/*
    CTHUGHA-L								sound.h
	
    Funktions that deal with sound.
	
*/

#ifndef __SOUND_H__
#define __SOUND_H__

#include "Option.h"

class OptionSound : public OptionInt {
public:
    OptionSound(const char * name, int v, int maxV = 0, int minV = 0) : 
	OptionInt(name, v, maxV, minV) {}
    void changeBy(int by);
};
	
class OptionChannels : public OptionSound {
public:
    OptionChannels(const char * name, int v) : 
	OptionSound(name, v, 2, 1) {}	//  1 or 2

    void changeBy(int by) {
	if(by != 0)
	    value = (value == 1) ? 2 : 1;
	OptionSound::changeBy(0);
    }
    const char* text() const {
	switch(value) {
	case 1:
	    return "mono";
	case 2:
	    return "stereo";
	default:
	    return "unknown";
	}
    }
};

extern OptionSound sound_method;
extern OptionSound sound_sample_rate;
extern OptionChannels sound_channels;
extern OptionSound sound_fragments;
extern OptionSound sound_fragment_size;
extern OptionSound sound_sample_size;
extern OptionOnOff sound_sync;


/* 
 *  General stuff
 */
int init_sound();
int exit_sound();

typedef char char2[2];
extern char2 * sound_data_unproc;
extern char2 * sound_data;			/* Stereo-Data */
extern int sound_bsize;
extern OptionOnOff sound_fork;			/* read in an extra process */
int sound_communicate(int to_child = 1);

extern int sine[320];

int sound_convert(unsigned char * data, int stereo, int bps);

/* 
 *  get-sound
 */
int get_sound();
extern OptionInt sound_minnoise;		/* quiet is below this */
extern OptionTime sound_quiet_change;		/* change after quiet-pause */
extern OptionTime sound_wait_quiet;		/* max. quiet interval */
extern OptionTime sound_wait_min;		/* min time between change */
extern OptionTime sound_wait_random;		/* extra random time */
extern OptionInt sound_firelevel;		/* level to change on beats */
extern int sound_wait_beat;			/* peaks till change */
extern int sound_attack;			/* attack level */
extern int sound_fire;				/* fired noew */
extern int sound_amplitude;			/* sound amplitude (variance) */
int sound_read();
int sound_fork_process();
int sound_kill_process();

/* 
 *  sound-processing
 */
int massage_audio();				/* audio-processing 1 */

int init_sound_process();
int sound_noise_check();


/*
 * things for the mixer
 */
int mixer_initial_volume(char * name, int volume);
int init_mixer();

extern char dev_mixer[];

/*
 * functions and variables for reading from the different sound devices
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

class SoundDevice {
protected:
    int error;
public:
    SoundDevice() : error(0) {}
    virtual ~SoundDevice() {}

    virtual void read() {};
    virtual void update() {};
    int err() const {
	return error;
    }
};

// interface to /dev/dsp
class SoundDeviceDSP : public SoundDevice {
protected:
    int handle;
    int mode;
    
    void setFragment();
    void setChannels();
    void setSampleRate();
    void setFormat();

    SoundDeviceDSP() {}		// only use one of the subclasses
    void init(int mode);
public:
    static char dev_dsp[];

    SoundDeviceDSP(int mode);
    virtual ~SoundDeviceDSP();

    virtual void update();
};

// /dev/dsp only for reading
class SoundDeviceDSPIn : public SoundDeviceDSP {
public:
    SoundDeviceDSPIn() : SoundDeviceDSP(O_RDONLY) {}
    virtual void read();
};

// /dev/dsp only for writing
class SoundDeviceDSPOut : public SoundDeviceDSP {
public:
    SoundDeviceDSPOut() : SoundDeviceDSP(O_WRONLY) {}
    
    void write(const unsigned char * buffer, int size);
};

// sound via network
class SoundDeviceNet : public SoundDevice {
protected:
    int handle;
    void net_request(int);
public:
    static char sound_hostname[256];
    
    SoundDeviceNet();
    virtual ~SoundDeviceNet();

    virtual void read();
    virtual void update();
};

// only random noise
class SoundDeviceRandom : public SoundDevice {
public:
    SoundDeviceRandom() : SoundDevice() {}
    virtual void read();
};

// get sound from a file
class SoundDeviceFile : public SoundDevice {
protected:
    FILE * file;
    SoundDeviceDSPOut * dsp;
    int buffer_pid;
 
    int openBuffer();
    int closeBuffer();

    virtual int open();
    virtual int close();
    virtual int header();
    virtual int restart();
    SoundDeviceFile() : SoundDevice(), file(NULL), dsp(NULL), buffer_pid(-1) {}
public:
    static char name[];

    SoundDeviceFile(int loud);
    virtual ~SoundDeviceFile();

    virtual void read();
    virtual void update();
};

// get sound from a program
class SoundDeviceExec : public SoundDeviceFile {
protected:
    virtual int open();
    virtual int close();
    virtual int restart();
    int childPid;
public:
    static char fifo[];

    SoundDeviceExec(int loud);
    virtual ~SoundDeviceExec();
};


// get sound data from child process
class SoundDeviceFork : public SoundDevice {
    int sound_child;
    int sound_shm_key;
    void * sound_shared;
public:
    SoundDeviceFork();
    virtual ~SoundDeviceFork();

    virtual void read();
    virtual void update();
};

extern SoundDevice * soundDevice;		// current sound device

enum SoundDeviceNr {
    SDN_DSPIn, SDN_Net, SDN_Random, SDN_File, SDN_FileLoud, 
    SDN_Exec, SDN_ExecLoud, SDN_Fork, SDN_Max
};
int newSoundDevice(SoundDeviceNr sDev);		// set a new sound device

extern OptionInt sound;				// which sound device to use
extern OptionSound soundBuffer;			// size of sound buffer (0 to disable it)
extern OptionOnOff soundLoop;

#endif



