#include "cthugha.h"
#include "translate.h"
#include "display.h"
#include "interface.h"
#include "tab_header.h"
#include "cth_buffer.h"

#include <unistd.h>

OptionOnOff use_translates("use-translate", 1);	/* allow translations */
OptionOnOff trans_stretch("stretching", 1);	/* allow stretching */

int done_translate=0;				/* trans. already done for
						   this iteration */


OptionListEntry * noneTrans = new TranslateEntry("none", "No Translate");
OptionOnOff * noneTransUse = new OptionOnOff("", 1);

OptionList translate("translate", &noneTrans, &noneTransUse, 1);

char translate_first[256];			/* first translate */
char lib_size[512];
static const char * translate_path[] = {
    "./",
    "./tab/",
    lib_size,
    CTH_LIBDIR "/tab/",
    ""
};

OptionListEntry * read_trans(FILE * file, char * name, char * dir);
OptionListEntry * read_cmdtrans(FILE * file, char * name, char * dir);

Interface * interfaceTranslate;
static InterfaceElement * interfaceTranslateElements[] = {
    new InterfaceElementTitle("Translate"),
    new InterfaceElementText("---------------------------------------"),
};
static int nInterfaceTranslateElements = sizeof(interfaceTranslateElements)/sizeof(InterfaceElement*);

/*
 * Initialize the translate-tables.
 */
int init_translate() {

    if ( int(use_translates))  {

	printfv(2,"  loading translates...\n");
	
	sprintf(lib_size, CTH_LIBDIR "/tab/%dx%d/", BUFF_WIDTH,BUFF_HEIGHT);
	
	translate.load(translate_path, "/tab/", ".tab", read_trans);
	translate.load(translate_path, "/tab/", ".cmd", read_cmdtrans);
	
	printfv(2,"  number of loaded translates: %d\n", translate.getNEntries() );
    }

    interfaceTranslate = new InterfaceList(interfaceTranslateElements, nInterfaceTranslateElements, &translate);

    return 0;
}


//
// Read a translate file
// 
static int * read_trans_data(FILE * file, const tab_header & header, 
			     int BSize, const char * name) {
    union data {
	unsigned long * l;
	unsigned short * s;
    } D;
    int i,j;

    int size = (BSize > 65535) ? sizeof(long) : sizeof(short);
    
    int * trans = new int[BSize];
    int * dst = trans;

    /* allocate memory for read buffer */
    D.l = new unsigned long[header.size_x];
    if ((void*)D.l != (void*)D.s) {
        printfe("Wackiness afoot at %d in %s\n",__LINE__,__FILE__);
        exit(1);
    }

    /* read data */
    for(i=0; i < header.size_y; i++) {
	if ( (j=fread( D.l, size, header.size_x, file)) < header.size_x ) {
	    printfe("  Can't read at line: %d, read: %d (%s)\n", i,j, name);
	    delete [] D.l;
	    delete [] trans;
	    return NULL;
	}	
	for(j=0; j < header.size_x; j++) {
	    if( BSize > 65535) {
		if ( D.l[j] >= (unsigned int)(BSize) ) {
		    printfe("  High-translate (value: %ld) in %s.\n", D.l[j], name);
		    delete [] D.l;
		    delete [] trans;
		    return NULL;
		}
		*dst++ = (int)D.l[j];
	    } else {
		if ( D.s[j] >= (unsigned int)(BSize) ) {
		    printfe("  High-translate (value: %d) in %s.\n",  D.s[j], name);
		    delete [] D.l;
		    delete [] trans;
		    return NULL;
		}
		*dst++ = (int)D.s[j];
	    }
	}
    }

    delete [] D.l;

    /* Check for too much data */
    if ( fread( &D, size, 1, file) ) {
	printfe("  Extra data at end of file %s\n", name);
	delete [] trans;
	return NULL;
    }

    return trans;
}

static int * stretch_trans(const int * src, const tab_header & header) {
    double xs,ys;
    int x,y,tp,ox,oy,dx,dy;
    int i,j;
    
    printfv(3," ... stretching");
   
    int * dst = new int[BUFF_HEIGHT*BUFF_WIDTH];
    
    xs = (double)header.size_x / BUFF_WIDTH;
    ys = (double)header.size_y / BUFF_HEIGHT;
    
    for (j=0;j<BUFF_HEIGHT;j++) {
	
	y = int( (double)j*ys );
	if (y >= header.size_y)
	    y = header.size_y - 1;
	
	for (i=0;i<BUFF_WIDTH;i++) {
	    
	    x = int( (double)i*xs );
	    if (x >= header.size_x)
		x = header.size_x - 1;
	    
	    tp = src[x + y*header.size_x];
	    ox = tp % header.size_x;
	    oy = tp / header.size_x;
	    dx = int( (double)(ox - x)/xs );
	    dy = int( (double)(oy - y)/ys );
	    dst[i+j*BUFF_WIDTH] = abs(i + dx + (j + dy) * BUFF_WIDTH)%BUFF_SIZE;
 	}
    }
    return dst;
}

OptionListEntry * read_trans(FILE * file, const char * name, const char * /*dir*/) {
    tab_header header;

    int BSize = BUFF_SIZE;
    int stretch = 0;
    TranslateEntry * new_trans;

    /* read header */
    if ( fread( &header, sizeof(tab_header), 1, file) != 1) {
	printfe("  Can't read header in file %s.\n", name);
	return NULL;
    }

    /* check header ID */
    if ( header.id != *((long*)"HDKB")) {
	printfv(2,"\n  Header ID-mismatch. Trying without header.");

	rewind(file);		// back to start of file

	/* fill in header */
	header.size_x = BUFF_WIDTH;
	header.size_y = BUFF_HEIGHT;

	new_trans = new TranslateEntry(name, "");
    } else {
	/* ID OK - now test size */
	if( (header.size_x != BUFF_WIDTH) || (header.size_y != BUFF_HEIGHT)) {
	    printfv(3,"\n    Size mismatch (%dx%d instead of %dx%d)", 
		    header.size_x, header.size_y, BUFF_WIDTH, BUFF_HEIGHT);
	    if( int(trans_stretch) ) {		/* allow stretching */
		stretch = 1;
	    } else {
		return NULL;
	    }
	}
	   
	/* translate table might use a different size */
	BSize = header.size_x * header.size_y;

	// make sure there is no \n in the name
	char * ent = strchr(header.description, '\n');
	if( ent != NULL) 
	    *ent = '\0';
	    
	new_trans = new TranslateEntry(name, header.description);
    }

    /* read data */
    int * trans = read_trans_data(file, header, BSize, name);
    if(trans == NULL) {
	delete new_trans;
	return NULL;
    }

    /* do the stretching if necessary (from: Rus Maxham) */
    if(stretch) {
	new_trans->trans = stretch_trans(trans, header);
	delete [] trans;
    } else
	new_trans->trans = trans;

    return new_trans;
}			


/*
 * read a file in the form:
 *  cmdtab
 *  <descrption>
 *  <commd> [args] %d %d
 */
#define MAX_DESC_LEN  64
OptionListEntry * read_cmdtrans(FILE * file, const char * name, const char * dir) {
    char line[512];
    char command[1024];
    FILE * cmd_file;
    int * dst;
    int i;
    TranslateEntry * new_trans;

    /* check ID */
    fgets(line, 512, file);
    if ( strncmp(line, "cmdtab", 6) != 0) {
	printfe("  Not a command translate file: %s.\n", name);
	return NULL;
    }

    // get description
    fgets(line, MAX_DESC_LEN, file);
    line[MAX_DESC_LEN-1] = '\0';	// make sure the string terminates
    // make sure there is no \n in the name
    char * ent = strchr(line, '\n');
    if( ent != NULL)
	*ent = '\0';
    
    new_trans = new TranslateEntry(name, line);

    /* get command */
    fgets(line, 512, file);

    /* generate the command line */
    if(strchr(line, '/') == NULL) {
	char cmd2[1024];
	strncpy(cmd2, dir, 1024);
	strncat(cmd2, "/", 1024);
	strncat(cmd2, line, 1024);
	sprintf(command, cmd2, BUFF_WIDTH, BUFF_HEIGHT);
    } else {
	sprintf(command, line, BUFF_WIDTH, BUFF_HEIGHT);
    }
    if(strlen(command) > 0)
	command[strlen(command) - 1] = '\0';

    /* start the command */
    printfv(3, "\n    starting: %s", command);
    if( (cmd_file = popen(command, "r")) == NULL) {
	printfee("  Can't run command '%s'.\n", command);
	return NULL;
    }

    new_trans->trans = new int[BUFF_SIZE];
    dst = (int*)new_trans->trans;
	
    /* read data from translate table program */
    for(i=0; i < BUFF_SIZE; i++, dst ++) {
	fread(dst, sizeof(long), 1, cmd_file);
	if( (*dst >= BUFF_SIZE) || (*dst < 0) ) {
	    printfe("  illegal translate (value: %d) in %s.\n", *dst, name);
	    pclose(cmd_file);
	    delete new_trans;
	    return NULL;
	}
    }
    if(fread(&i, 1, 1, cmd_file)) {		/* something more to read */
	printfe("  Extra data at end of table in %s.\n", name);
	pclose(cmd_file);
	delete new_trans;
	return NULL;
    }

    pclose(cmd_file);

    return new_trans;
}



//
// Do the translate
//
int TranslateEntry::operator()() {
    int i;
    unsigned int * dst; 
    unsigned char * src;

    if( done_translate)
	return 0;

    const int * trans = this->trans;
    if(trans == NULL)
	return 0;

    dst = (unsigned int *)passive_buffer;
    src = active_buffer;
    active_buffer = passive_buffer; passive_buffer = src;
			
    src[0] = 0;

    /* always write 4 values at once */
    for(i=BUFF_SIZE/4; i != 0; i--) {  		/* do the translate */
	unsigned int D = src[*trans++];
	D <<= 8;
	D = D + src[*trans++];
	D <<= 8;
	D = D + src[*trans++];
	D <<= 8;
	*dst++ = D + src[*trans++];
    }
    return 0;
}












